"""Edge extension for dataframe"""

from pandas import DataFrame, Series
class EdgeDataFrame(DataFrame): 
    @property
    def _constructor(self): 
        return EdgeDataFrame 
      
    @property
    def _constructor_sliced(self): 
        return EdgeSeries 

        
class EdgeSeries(Series): 
    @property
    def _constructor(self): 
        return EdgeSeries 
      
    @property
    def _constructor_sliced(self): 
        return EdgeSeries 


"""Calculate communities"""

from typing import Union
import graph_tool.all as gt
import numpy as np
import pandas as pd

from .nodedataframe import NodeDataFrame
from .state import state_unnest


def group_sbm(G: gt.Graph, deg_corr: bool = True) -> NodeDataFrame:
  """Run degree-corrected Stochastic Block Model.

  See the `stochastic block model` section in graph-tool HOWTO. 

  :return: Node dataframe with group membership, model entropy, and state object.
  """
  state = gt.minimize_blockmodel_dl(G, deg_corr=deg_corr)
  ent = state.entropy()
  blocks = list(state.get_blocks())
  return NodeDataFrame({"group": [blocks], "ent": ent, "state_obj": state})


def group_sbm_covar(G: gt.Graph, 
                    edge_covar: gt.EdgePropertyMap, 
                    family: str,
                    deg_corr: bool = True) -> NodeDataFrame:
  """Run Stochastic Block Model with weight models.
  
  See `edge-weights-and-covariates` section in graph-tool HOWTO.

  :return: Node dataframe with list-columns for each level, and the model entropy
  """
  if family in ['real-exponential', "real-normal"]:
    assert all([type(v) is float for v in list(G.ep[f"{edge_covar}"])])
  elif family in ['discrete-geometric', 'discrete-binomial', 'discrete-poisson']:
    assert all([type(v) is int for v in list(G.ep[f"{edge_covar}"])])
  
  state = gt.minimize_blockmodel_dl(
    G,
    state_args=dict(recs=[G.ep[f"{edge_covar}"]], rec_types=[f"{family}"]), 
    deg_corr=deg_corr
  )

  ent = state.entropy()
  blocks = list(state.get_blocks())
  
  return NodeDataFrame({"group": [blocks], "ent": ent, "state_obj": state})


def group_sbm_marginal(G: gt.Graph) -> NodeDataFrame:
  """Model averaging on SBM by sampling partitions of the posterior.

  EXAMPLE
  =======
  g = gt.collection.data["lesmis"]
  group_marginals = group_sbm_marginal(g)
  state = group_marginals.state_obj[0]
  pv = group_marginals.marginals[0]
  state.draw(pos=g.vp.pos, vertex_shape="pie", vertex_pie_fractions=pv)

  See `inferring-the-best-partition` section in graph-tool HOWTO.

  :return: Node dataframe with list-columns for each level, 
           model entropy, state object, and the marginals
  """
  state = gt.BlockState(G)

  # Equilibrate markov chain
  gt.mcmc_equilibrate(state, wait=1000, mcmc_args=dict(niter=10))
  
  bs = []
  def collect_partitions(s):
    nonlocal bs
    bs.append(s.b.a.copy())

  # Collect partitions for exactly 100K sweeps, at intervals of 10 sweeps
  gt.mcmc_equilibrate(state, force_niter=10000, mcmc_args=dict(niter=10),
                      callback=collect_partitions)

  # Disambiguate partitions and obtain marginals
  pmode = gt.PartitionModeState(bs, converge=True)
  pv = pmode.get_marginal(G)
  ent = state.entropy()
  blocks = list(state.get_blocks())

  return NodeDataFrame({"group": [blocks], "marginals": [pv], 
                        "ent": ent, "state_obj": state})


def group_measured(G: gt.Graph, 
                   n: gt.EdgePropertyMap, 
                   x: gt.EdgePropertyMap,
                   n_default: int = 1,
                   x_default: int = 0):
  """Model averaging on SBM by sampling partitions of the posterior.

  EXAMPLE
  =======
  g = gt.collection.data["lesmis"]
  group_marginals = group_sbm_marginal(g)
  state = group_marginals.state_obj[0]
  pv = group_marginals.marginals[0]
  state.draw(pos=g.vp.pos, vertex_shape="pie", vertex_pie_fractions=pv)

  See `Measured networks` section in graph-tool HOWTO.

  :return: Node dataframe with list-columns for each level, 
           model entropy, state object, and the marginals
  """
  state = gt.MeasuredBlockState(G, n=n, n_default=n_default, x=x, x_default=x_default)

  gt.mcmc_equilibrate(state, wait=100, mcmc_args=dict(niter=10))
  
  u = None              # marginal posterior edge probabilities
  bs = []               # partitions
  cs = []               # average local clustering coefficient
  def collect_partitions(s):
    nonlocal bs
    bs.append(s.b.a.copy())

  def collect_marginals(s):
    nonlocal u, bs, cs
    u = s.collect_marginal(u)
    bstate = s.get_block_state()
    bs.append(bstate.levels[0].b.a.copy())
    cs.append(gt.local_clustering(s.get_graph()).fa.mean())

  gt.mcmc_equilibrate(state, force_niter=10000, mcmc_args=dict(niter=10),
                      callback=collect_marginals)

  ent = state.entropy()
  #!TODO: check what kind of object that bstate returns. probably a nested one.
  bstate = state.get_block_state()

  return NodeDataFrame({"edge_prob_ob": [u], "ent": ent, "state_obj": state})


def group_sbm_assortative(G: gt.Graph) -> NodeDataFrame:
  """Run planted partition model.

  See `Assortative community structure` section in graph-tool HOWTO.

  :return: Node dataframe with list-columns for each level, 
           model entropy, state object, and the marginals
  """
  state = gt.PPBlockState(G)

  # 1,000 sweeps of the MCMC with zero temperature
  state.multiflip_mcmc_sweep(beta=np.inf, niter=1000)

  ent = state.entropy()
  blocks = list(state.get_blocks())

  return NodeDataFrame({"group": [blocks], "ent": ent, "state_obj": state})


def group_hsbm(G: gt.Graph, 
               deg_corr: bool = True, 
               merge_split: bool = True) -> NodeDataFrame:
  """Run hierarchical degree-corrected Stochastic Block Model.

  See the `nested stochastic block model` section in graph-tool HOWTO.
  
  :return: node dataframe with a column for each level, model entropy, and state object.
  """
  state = gt.minimize_nested_blockmodel_dl(G, deg_corr=deg_corr)

  if merge_split == True:
    state = state.copy(bs=state.get_bs() + [np.zeros(1)] * 4, sampling=True)

    for i in range(100):
      ret = state.multiflip_mcmc_sweep(niter=10, beta=np.inf)

  group_all_levels = state_unnest(state)
  group_all_levels = pd.concat(
    [NodeDataFrame({f"{col}": [group_all_levels[f"{col}"]]}) 
     for col in group_all_levels.columns], axis=1
  )
  return pd.concat(
    [group_all_levels, NodeDataFrame({"ent": [ent], "state_obj": state})], 
     axis=1)


def group_hsbm_covar(G: gt.Graph, 
                     edge_covar, 
                     family: str,
                     deg_corr: bool = True, 
                     merge_split: bool = True) -> NodeDataFrame:
  """Run hierarchical Stochastic Block Model with weight models
  
  See the `Edge weights and covariates` section in graph-tool HOWTO.

  :return: node dataframe with a column for each level, model entropy, and state object.
  """
  if family in ['real-exponential', "real-normal"]:
    assert all([type(v) is float for v in list(edge_covar)])
  elif family in ['discrete-geometric', 'discrete-binomial', 'discrete-poisson']:
    assert all([type(v) is int for v in list(edge_covar)])
  
  state = gt.minimize_nested_blockmodel_dl(
    G, 
    state_args=dict(recs=[edge_covar], rec_types=[f"{family}"])
  )

  if merge_split == True:
    state = state.copy(bs=state.get_bs() + [np.zeros(1)] * 4, sampling=True)

    for i in range(100):
      ret = state.multiflip_mcmc_sweep(niter=10, beta=np.inf)
  
  ent = state.entropy()
  group_all_levels = state_unnest(state)
  group_all_levels = pd.concat(
    [NodeDataFrame({f"{col}": [group_all_levels[f"{col}"]]}) 
     for col in group_all_levels.columns], axis=1
  )
  return pd.concat([group_all_levels, NodeDataFrame({"ent": [ent], "state_obj": state})], axis=1)


def group_hsbm_layered(G: gt.Graph, 
                       edge_covar, 
                       merge_split: bool = True) -> NodeDataFrame:
  """Run hierarchical Stochastic Block Model with weight models.
  
  EXAMPLE
  =======
  g = gt.collection.ns["new_guinea_tribes"]
  group_layered = group_hsbm_layered(g, g.ep.weight, merge_split=False)
  state = group_layered.state_obj[0]
  state.draw(edge_color=g.ep.weight, edge_gradient=[],
             ecmap=(plt.cm.coolwarm_r, .6), edge_pen_width=5)

  See `Layered networks` section in graph-tool HOWTO.

  :return: node dataframe with a column for each level, model entropy, and state object.
  """

  state = gt.minimize_nested_blockmodel_dl(
    G, 
    state_args=dict(ec=G.ep.weight, layers=True),
    layers=True
  )

  if merge_split == True:
    state = state.copy(bs=state.get_bs() + [np.zeros(1)] * 4, sampling=True)

    for i in range(100):
      ret = state.multiflip_mcmc_sweep(niter=10, beta=np.inf)
  
  ent = state.entropy()
  group_all_levels = state_unnest(state)
  group_all_levels = pd.concat(
    [NodeDataFrame({f"{col}": [group_all_levels[f"{col}"]]}) 
     for col in group_all_levels.columns], axis=1
  )
  return pd.concat([group_all_levels, NodeDataFrame({"ent": [ent], "state_obj": state})], axis=1)


def group_hsbm_marginal(G):
  """Model averaging on SBM by sampling partitions of the posterior.

  EXAMPLE
  =======
  g = gt.collection.data["lesmis"]
  group_marginals = group_hsbm_marginal(g)
  state = group_marginals.state_obj[0]
  pv = group_marginals.marginals[0]
  state.draw(pos=g.vp.pos, vertex_shape="pie", vertex_pie_fractions=pv)

  See `Hierarchical partitions` section in graph-tool HOWTO.

  :return: Node dataframe with list-columns for each level, 
           model entropy, state object, and the marginals
  """
  state = gt.NestedBlockState(G)

  # Equilibrate markov chain
  gt.mcmc_equilibrate(state, wait=1000, mcmc_args=dict(niter=10))
  
  bs = []
  def collect_partitions(s):
    nonlocal bs
    bs.append(s.get_bs())

  # Collect partitions for exactly 100K sweeps, at intervals of 10 sweeps
  gt.mcmc_equilibrate(state, force_niter=10000, mcmc_args=dict(niter=10),
                      callback=collect_partitions)

  # Disambiguate partitions and obtain marginals
  pmode = gt.PartitionModeState(bs, nested=True, converge=True)
  pv = pmode.get_marginal(G)

  # Get consensus estimate
  bs = pmode.get_max_nested()

  state = state.copy(bs=bs)

  group_all_levels = state_unnest(state)
  group_all_levels = pd.concat(
    [NodeDataFrame({f"{col}": [group_all_levels[f"{col}"]]}) 
     for col in group_all_levels.columns], axis=1
  )

  ent = state.entropy()
  return pd.concat(
    [group_all_levels, NodeDataFrame({"marginals": [pv], "ent": [ent], "state_obj": state})], 
     axis=1)


def group_latent_poisson(G: gt.Graph):
  """Model averaging on SBM by sampling partitions of the posterior.

  EXAMPLE
  =======
  g = gt.collection.data["polbooks"]
  group_poisson = group_latent_poisson(g)
  bstate = group_poisson.state_obj[0]
  u = group_poisson.u[0]
  pv = group_poisson.marginals[0]
  ew = group_poisson.ew[0]
  bstate.draw(pos=u.own_property(g.vp.pos), vertex_shape="pie", vertex_pie_fractions=pv,
              edge_pen_width=gt.prop_to_size(ew, .1, 8, power=1), edge_gradient=None)

  See `Latent Poisson multigraphs¶` section in graph-tool HOWTO.

  :return: Node dataframe with list-columns for each level, 
           model entropy, state object, and the marginals
  """
  state = gt.LatentMultigraphBlockState(G)

  gt.mcmc_equilibrate(state, wait=100, mcmc_args=dict(niter=10))
  
  u = None   # marginal posterior multigraph
  bs = []    # partitions
  def collect_marginals(s):
    nonlocal bs, u
    u = s.collect_marginal_multigraph(u)
    bstate = state.get_block_state()
    bs.append(bstate.levels[0].b.a.copy())


  gt.mcmc_equilibrate(state, force_niter=10000, mcmc_args=dict(niter=10),
                      callback=collect_marginals)

  # compute average multiplicities
  ew = u.new_ep("double")
  w = u.ep.w
  wcount = u.ep.wcount
  for e in u.edges():
      ew[e] = (wcount[e].a * w[e].a).sum() / wcount[e].a.sum()

  bstate = state.get_block_state()
  bstate = bstate.levels[0].copy(g=u)

  # Disambiguate partitions and obtain marginals
  pmode = gt.PartitionModeState(bs, converge=True)
  pv = pmode.get_marginal(u)

  blocks = list(bstate.get_blocks())
  ent = bstate.entropy()
  return NodeDataFrame({"group": [blocks], "marginals": [pv], "ent": [ent], 
                       "ew": [ew], "u":[u], "state_obj": bstate})


# minimize_blockmodel_dl()        --> BlockState 
# minimize_nested_blockmodel_dl() --> NestedBlockState

# state.mcmc_sweep                -->
#   - single node is moved at a time
# state.multiflip_mcmc_sweep      -->
#   - merge and splits
# gt.mcmc_anneal(state, ...)      -->
# gt.mcmc_equilibrate()           -->
# gt.PPBlockState                 -->
# MeasuredBlockState
# MixedMeasuredBlockState
# UncertainBlockState
# LatentMultigraphBlockState
# EpidemicsBlockState


"""Miscellaneous internal tidygraphtool helper functions."""

from .as_data_frame import as_data_frame
from typing import (
    Iterable,
    Union,
    Hashable,
    Optional
)

import re
import graph_tool.all as gt
import numpy as np
import pandas as pd
import pandas_flavor as pf


def assert_nodes_edges_equal(nodes, edges, node_key):
    check_column(nodes, [f"{node_key}"])

    x1 = nodes[f"{node_key}"].astype(str)
    x2 = _extract_nodes(edges)["name"]

    assert set(x1) == set(x2)


def assert_nodes_nodes_equal(G, nodes_df):
    nb_nodes_g = len(list(G.vertices()))
    nb_nodes_df = len(nodes_df)
    assert nb_nodes_g == nb_nodes_df


def assert_edges_edges_equal(G, edges_df):
    nb_edges_g = len(list(G.edges()))
    nb_edges_df = len(edges_df)
    assert nb_edges_g == nb_edges_df


def assert_index_reset(df):
    assert all(df.index == range(len(df)))


def assert_edges_indexified(nodes, edges):
    assert set(edges.source).union(set(edges.target)) == set(nodes.index)


def assert_nodes_mergeable(nodes, edges, left_key, right_key):
    assert all(edges[f"{left_key}"].isin(nodes[f"{right_key}"]))


def check(varname: str, value, expected_types: list):
    """
    One-liner syntactic sugar for checking types.

    :param varname: The name of the variable.
    :param value: The value of the varname.
    :param expected_types: The types we expect the item to be.
    :raises TypeError: if data is not the expected type.
    """
    is_expected_type: bool = False
    for t in expected_types:
        if t is callable:
            is_expected_type = t(value)
        else:
            is_expected_type = isinstance(value, t)
        if is_expected_type:
            break

    if not is_expected_type:
        raise TypeError(
            "{varname} should be one of {expected_types}".format(
                varname=varname, expected_types=expected_types
            )
        )


def check_args_order(x, y, z):
    assert isinstance(x, gt.Graph)
    assert isinstance(y, str)
    assert callable(z)


def check_column(
    df: pd.DataFrame, column_names: Union[Iterable, str], present: bool = True
):
    """
    One-liner syntactic sugar for checking the presence or absence of columns.
    
    :param df: The name of the variable.
    :param column_names: A list of column names we want to check to see if
        present (or absent) in df.
    :param present: If True (default), checks to see if all of column_names
        are in df.columns. If False, checks that none of column_names are
        in df.columns.
    :raises ValueError: if data is not the expected type.
    """
    for column_name in column_names:
        if present and column_name not in df.columns:  # skipcq: PYL-R1720
            raise ValueError(
                f"{column_name} not present in dataframe columns!"
            )
        elif not present and column_name in df.columns:
            raise ValueError(
                f"{column_name} already present in dataframe columns!"
            )


def check_null_values(df: pd.DataFrame,  column_names: Union[Iterable, str]):
    """
    One-liner syntactic sugar for checking the null values of columns.
    :raises ValueError: if data is null values.
    """
    check_column(df, column_names)
    assert any([df[f'{colname}'].isnull().values.any() == True
                for colname
                in column_names]) == False


def _extract_nodes(edges: pd.DataFrame) -> pd.DataFrame:
    # We loose order here. Issues when we first create graph with
    # edgelist, then we want to augment it.
    all_users = set(edges.iloc[:, 0]).union(set(edges.iloc[:, 1]))
    return pd.DataFrame(all_users).rename(columns={0: "name"}).astype(str)


def _convert_types2gt(df, prop_name):
    check_column(df, [prop_name], present=True)
    prop_type = str(df[f"{prop_name}"].dtype)

    if prop_type == "object":
        prop_type = "string"
    elif prop_type == "float64":
        prop_type = "float"
    elif re.match(r"^int", prop_type):
        prop_type = prop_type + "_t"
    elif prop_type == 'bool':
        prop_type = 'bool'
    else:
        raise ValueError("Failed to guess type")

    return prop_type


def _find_namecol(x, nodes):
    if isinstance(x, gt.Graph):
        namecol = set(list(x.vp.name))
    elif guess_df_type(x) == "EdgeDataFrame":
        namecol = set(x.source).union(set(x.target))
    else:
        raise ValueError("Unable to guess x")
    return [c for c in nodes.columns
            if set(nodes[f"{c}"]) == namecol][0]


def guess_df_type(x: pd.DataFrame) -> str:
    x = pd.DataFrame(x)
    colnames = x.columns.str.lower()
    if any(colnames.isin(["source", "target", "from", "to", "weight"])):
        return "EdgeDataFrame"
    elif any(colnames.isin(["id", "name", "user"])):
        return "NodeDataFrame"
    else:
        raise ValueError("Unable to guess dataframe type")


def guess_dict_type(x: list) -> str:
    if len(x) == 2:
        x1, x2 = [_ for _ in x]
        if any([_ in ['nodes', 'vertices'] for _ in (x1, x2)]) & \
                any([_ in ['edges', 'links'] for _ in (x1, x2)]):
            return 'node_edge'
        else:
            raise ValueError("Unknowen list format")


@pf.register_dataframe_method
def reorder_columns(
    df: pd.DataFrame, column_order: Union[Iterable[str], pd.Index, Hashable]
) -> pd.DataFrame:
    """Reorder DataFrame columns by specifying desired order as list of col names.



    :param df: `DataFrame` to reorder
    :param column_order: A list of column names or Pandas `Index`
        specifying their order in the returned `DataFrame`.
    :returns: A pandas DataFrame with reordered columns.
    :raises IndexError: if a column within ``column_order`` is not found
        within the DataFrame.
    """
    check("column_order", column_order, [list, tuple, pd.Index])

    if any(col not in df.columns for col in column_order):
        raise IndexError(
            "A column in ``column_order`` was not found in the DataFrame."
        )

    # if column_order is a Pandas index, needs conversion to list:
    column_order = list(column_order)

    return df.reindex(
        columns=(
            column_order
            + [col for col in df.columns if col not in column_order]
        ),
        copy=False,
    )




"""Pipeable dataframe functions - Experimental"""

import pandas as pd
import numpy as np
from .utils import check
from typing import (
    Any,
    List,
    Tuple,
    Union,
)


def add_column(df: pd.DataFrame,
               column_name: str,
               value: Union[List[Any], Tuple[Any], Any],
               fill_remaining: bool = False) -> pd.DataFrame:
    df = df.copy()
    check("column_name", column_name, [str])

    if column_name in df.columns:
        raise ValueError(
            f"Attempted to add column that already exists: " f"{column_name}."
        )

    nrows = df.shape[0]

    if hasattr(value, "__len__") and not isinstance(
        value, (str, bytes, bytearray)
    ):

        if len(value) > nrows:
            raise ValueError(
                "`value` has more elements than number of rows "
                f"in your `DataFrame`. vals: {len(value)}, "
                f"df: {nrows}"
            )
        if len(value) != nrows and not fill_remaining:
            raise ValueError(
                "Attempted to add iterable of values with length"
                " not equal to number of DataFrame rows"
            )

        if len(value) == 0:
            raise ValueError(
                "`value` has to be an iterable of minimum length 1"
            )
        len_value = len(value)
    elif fill_remaining:
        # relevant if a scalar val was passed, yet fill_remaining == True
        len_value = 1
        value = [value]

    nrows = df.shape[0]

    if fill_remaining:
        times_to_loop = int(np.ceil(nrows / len_value))

        fill_values = list(value) * times_to_loop

        df[column_name] = fill_values[:nrows]
    else:
        df[column_name] = value

    return df




"""Calculate node and edge centrality"""

from graph_tool import VertexPropertyMap
import pandas as pd
import graph_tool.all as gt 
from typing import Optional

from .nodedataframe import NodeDataFrame
from .edgedataframe import EdgeDataFrame
from .as_data_frame import as_data_frame
from .context import expect_nodes, expect_edges


def centrality_betweenness(G: gt.Graph) -> pd.Series:
    expect_nodes(G)
    bet, _ = gt.betweenness(G)
    return NodeDataFrame({"bet": list(bet)})["bet"]


def centrality_betweenness(G: gt.Graph) -> pd.Series:
    expect_nodes(G)
    bet, _ = gt.betweenness(G)
    return NodeDataFrame({"bet": list(bet)})["bet"]


def centrality_edge_betweenness(G: gt.Graph) -> pd.Series:
    expect_edges(G)
    _, bet = gt.betweenness(G)
    return EdgeDataFrame({"bet": list(bet)})["bet"]


def centrality_closeness(G: gt.Graph) -> pd.Series:
    expect_nodes(G)
    return NodeDataFrame({"closeness": list(gt.closeness(G))})["closeness"]


def centrality_degree(G: gt.Graph,
                      weight=None,
                      mode='out') -> pd.Series:
    expect_nodes(G)
    deg = G.degree_property_map(f"{mode}", weight=weight)
    return NodeDataFrame({"deg": list(deg)})["deg"]


def centrality_eigenvector(G: gt.Graph,
                           weight: Optional[float] = None,
                           vprop: Optional[gt.VertexPropertyMap] = None,
                           epsilon: float = 1e-06,
                           max_iter: int = None) -> pd.Series:
    """
    Calculate the eigenvector centrality of each vertex in the graph.

    Wrapper for [graph_tool.centrality.eigenvector()]. See ?gt.eigenvector

    Functional usage syntax with pipes:
    
    .. code-block:: python
    
        np.random.seed(42)
        import matplotlib.pyplot as plt
            
        g = (gt.collection.data["polblogs"] >> 
              activate("nodes") >>
              filter_largest_component())

        g = (g >> 
              activate("edges") >>
              augment_prop(w, prop_name="w") >>
              activate("nodes") >>
              add_property("x", centrality_eigenvector))

        gt.graph_draw(g, pos=g.vp["pos"], vertex_fill_color=g.vp.x,
                      vertex_size=gt.prop_to_size(g.vp.x, mi=5, ma=15),
                      vcmap=plt.cm.gist_heat,
                      vorder=g.vp.x)

    :returns: A Node Series containing the eigenvector values.
    """
    expect_nodes(G)
    _, eigen_vector = gt.eigenvector(G, weight=weight, vprop=vprop,
                                     epsilon=epsilon,
                                     max_iter=max_iter)
    return NodeDataFrame({"eigen": list(eigen_vector)})["eigen"]


def centrality_eigentrust(G: gt.Graph,
                          trust_map: Optional[gt.EdgePropertyMap] = None,
                          vprop: Optional[gt.VertexPropertyMap] = None,
                          norm: bool = False,
                          epsilon: float = 1e-06,
                          max_iter: int = 0,
                          ret_iter: bool =False) -> pd.Series:
    """
    Calculate the eigentrust centrality of each vertex in the graph.

    Wrapper for [graph_tool.centrality.eigentrust()]. See ?gt.eigentrust

    Functional usage syntax with pipes:
    
    .. code-block:: python
      
      np.random.seed(42)
      import matplotlib.pyplot as plt
    
      g = (
        gt.collection.data["polblogs"] >> 
          activate("nodes") >>
          filter_largest_component()
      )
      w = EdgeDataFrame({"w":np.random.random(len(list(g.edges()))) * 42})
      g = (
        g >> 
          activate("edges") >> 
          augment_prop(w, prop_name="w") >> 
          activate("nodes")
      )
      g = augment_prop(g, centrality_eigentrust(g, g.ep.w), prop_name="t")
      gt.graph_draw(
          g, pos=g.vp["pos"], vertex_fill_color=g.vp.t,
          vertex_size=gt.prop_to_size(g.vp.t, mi=5, ma=15),
          vcmap=plt.cm.gist_heat,
          vorder=g.vp.t
      )


    :param trust_map: Edge property map with the values of trust associated with each edge. A float bounded between [0,1].
    :returns: A Node Series containing the eigentrust values.
    """
    expect_nodes(G)
    eigen_trust = gt.eigentrust(G, trust_map=trust_map, vprop=vprop, norm=norm, 
                                 epsilon=epsilon, max_iter=max_iter, ret_iter=ret_iter)
    return NodeDataFrame({"t": list(eigen_trust)})["t"]


def centrality_pagerank(G: gt.Graph) -> pd.Series:
    """
    Calculate the eigentrust centrality of each vertex in the graph.

    Wrapper for [graph_tool.centrality.pagerank()]. See ?gt.pagerank

    Functional usage syntax with pipes:
    
    .. code-block:: python

        g = (gt.collection.data["polblogs"] >> 
              activate("nodes") >>
              filter_largest_component() >>
              add_property("pr", centrality_pagerank))

        gt.graph_draw(g, pos=g.vp["pos"], vertex_fill_color=g.vp.pr,
                      vertex_size=gt.prop_to_size(g.vp.pr, mi=5, ma=15),
                      vorder=g.vp.pr, vcmap=plt.cm.gist_heat)

    """
    expect_nodes(G)
    return NodeDataFrame({"pr": list(gt.pagerank(G))})["pr"]




"""Helper functions to work with `state` object"""

import re
import pandas as pd
import numpy as np
import graph_tool.all as gt
from pipey import Pipeable


def state_marginals(G: gt.Graph, state) -> pd.DataFrame:
  """Get marginals of state object
  
  EXAMPLE
  =======
  import seaborn as sns
  import matplotlib.pyplot as plt

  g = gt.collection.data["lesmis"]
  group_marginals = group_sbm_marginal(g)
  d = get_marginal_probability(g, state)

  sns.set_style("whitegrid")
  sns.barplot(x="x", y="h", dodge = True, linewidth=0.5,
              data=d, palette="deep", edgecolor=".2")
  plt.xlabel("B")
  plt.ylabel("P(B|A)");
  
  See `inferring-the-best-partition` section in graph-tool HOWTO.

  :return: Dataframe with marginal probabilities of blocks
  """
  h = np.zeros(G.num_vertices() + 1)

  def collect_num_groups(s):
      B = s.get_nonempty_B()
      h[B] += 1

  gt.mcmc_equilibrate(state, force_niter=10000, 
                      mcmc_args=dict(niter=10),
                      callback=collect_num_groups)

  return pd.DataFrame({"h":h})\
           .query("h != 0")\
           .assign(h = lambda x: x.h / x.h.sum(),
                   B = lambda x: x.index)


def state_nested_marginals(G: gt.Graph, state) -> pd.DataFrame:
  """Get marginals of NestedBlockState object
  
  import seaborn as sns
  import matplotlib.pyplot as plt

  EXAMPLE
  =======
  g = gt.collection.data["lesmis"]
  group_marginals = group_sbm_marginal(g)
  d = get_marginal_probability(g, state)
  d_long = d.apply(pd.Series.explode)

  levels = d_long.levels.unique()[:-1] #trivially 1
  fig, ax = plt.subplots(1,len(levels),figsize=(25,5))

  for i, level in enumerate(levels):
    sns.barplot(x="B", y="h", data=d_long[d_long.levels == level],  ax=ax[i],
                dodge = True, linewidth=0.5,
                palette="deep", edgecolor=".2")
    ax[i].set_xlabel(f"$B_{level}$")
    ax[i].set_ylabel(f"$P(B_{level}|A)$")

  sns.set_style("whitegrid")
  sns.barplot(x="B", y="h", dodge = True, linewidth=0.5,
              data=d_long, palette="deep", edgecolor=".2")
  plt.xlabel("B")
  plt.ylabel("P(A|B)");
  
  See `Hierarchical partitions` section in graph-tool HOWTO.

  :return: Dataframe with marginal probabilities of blocks
  """
  h = [np.zeros(G.num_vertices() + 1) for s in state.get_levels()]

  def collect_num_groups(s):
      for l, sl in enumerate(s.get_levels()):
        B = sl.get_nonempty_B()
        h[l][B] += 1

  # Collect marginal for 100K sweeps
  gt.mcmc_equilibrate(state, force_niter=10000, mcmc_args=dict(niter=10),
                      callback=collect_num_groups)

  df_long = pd.DataFrame({"h":h}).explode("h")\
              .assign(levels = lambda x: x.index,
                      B = lambda x: x.groupby("levels").cumcount(+1))\
              .query("h != 0")\
              .assign(h = lambda x: x.h/x.h.sum())\
              .assign(h = lambda x: x.h.astype(float).round(decimals=3))
  
  nb_levels = len(df_long.levels.unique())

  return pd.DataFrame(
    {"h": [df_long.query(f"levels == {lvl}").h.tolist() for lvl in range(nb_levels)],
     "B": [df_long.query(f"levels == {lvl}").B.tolist() for lvl in range(nb_levels)]}
    ).assign(levels = lambda x: x.index)


def _merge_level_below(df_lvl_below, dat):
    lvl_below = list(df_lvl_below.columns)[-1]
    lvl_below = int(re.sub("hsbm_level", "", lvl_below))
    
    df_lvl_above = pd.DataFrame({f"hsbm_level{lvl_below+1}": dat})
    
    # index of df level i == block level i-1/agents
    df_lvl_above[f'hsbm_level{lvl_below}'] = df_lvl_above.index
    
    # We left join level i-1 of df_lvl_i onto level i-1 of df_lvl_be
    return pd.merge(df_lvl_below, df_lvl_above, 
                    left_on = f"hsbm_level{lvl_below}", 
                    right_on = f"hsbm_level{lvl_below}", 
                    how = "left")


@Pipeable(try_normal_call_first=True)
def state_unnest(state):
    """
    Unnest gt.BlockState object into a dataframe.
    
    The function will create one column by level.
    """
    if isinstance(state, gt.BlockState):
        raise ValueError("Not hierarchical block state")

    levels = state.get_levels()
    list_r_level = [list(r for r in levels[i].get_blocks()) for i in range(len(levels))]
    com_all_lvl = pd.DataFrame({"hsbm_level0": list_r_level[0]})
    
    i=1
    while (i < len(list_r_level)):
        com_all_lvl = _merge_level_below(com_all_lvl, list_r_level[i])
        i += 1
    
    return com_all_lvl



"""Main module to convert to get graph"""

from functools import singledispatch
from typing import Optional, Dict
from pipey import Pipeable

from .as_data_frame import as_data_frame
from .augment import augment_prop
from .utils import assert_nodes_edges_equal, assert_nodes_mergeable, check_column, guess_df_type, guess_dict_type, _extract_nodes
from .nodedataframe import NodeDataFrame, NodeSeries
from .edgedataframe import EdgeDataFrame, EdgeSeries
from .context import activate

import graph_tool.all as gt
import pandas as pd


def gt_graph(
    nodes: Optional[pd.DataFrame] = None,
    edges: Optional[pd.DataFrame] = None,
    directed: bool = True,
    node_key: str = "name"
) -> gt.Graph:
    """
    Creates a graph-tool object using nodes and edges dataframe.

    Example
    =======
    .. code-block:: python

        nodes = pd.DataFrame({"name": ["Bob", "Alice", "Joan", "Melvin"]}) 
        edges = pd.DataFrame({"source": ["Bob", "Bob", "Joan", "Alice"],
                              "target": ["Joan", "Melvin", "Alice", "Joan"]})
        gt_graph(nodes, edges)

    NOTE
    ====
    Params are adapted directly from thomasp85's tidygraph api in R 

    :param nodes: A `data.frame` containing information about the nodes in the graph.
        If `edges.target` and/or `edges.from` are characters then they will be
        matched to the column named according to `node_key` in nodes, if it exists.
        If not, they will be matched to the first column.
    :param edges:  A `data.frame` containing information about the edges in the
        graph. The terminal nodes of each edge must either be encoded in a `target` and
        `source` column, or in the two first columns, as integers. These integers refer to
        `nodes` index.
    :directed:  Should the constructed graph be directed (defaults to `TRUE`)
    :node_key: The name of the column in `nodes` that character represented
        `target` and `source` columns should be matched against. 
    :return: A `Graph_tool` object
    """
    return as_gt_graph({"nodes": nodes, "edges": edges}, 
                        directed=directed, 
                        node_key=node_key)


@singledispatch
def as_gt_graph(x):
    raise ValueError(f"{x} is not implemented")


@as_gt_graph.register(pd.DataFrame)
@as_gt_graph.register(EdgeDataFrame)
def _data_frame(x, directed=True, node_key='name') -> gt.Graph:
    """Convert dataframe, and returns a graph-tool graph."""
    if guess_df_type(x) == 'EdgeDataFrame':
        nodes = _extract_nodes(x)
        node_edge_df = {"nodes": nodes, "edges": x}
        nodes, edges = _as_graph_node_edge(node_edge_df, node_key=f'{node_key}')

        # Create Graph
        g = gt.Graph(directed=directed) >> activate("nodes")
        g.add_edge_list(edges[["source", "target"]].to_numpy())
        g = augment_prop(g, NodeDataFrame(nodes), prop_name=f"{node_key}")

        # # Add edge metadata
        g = activate(g, "edges")
        edgecols = edges.iloc[:, 2::].columns
        if len(edgecols) == 1:
            augment_prop(g, EdgeDataFrame(edges), prop_name=edgecols[0])
        elif len(edgecols) > 1:
            [augment_prop(g, EdgeDataFrame(edges), prop_name=c) for c in edgecols]

        return activate(g, "nodes") # nodes by default
    else:
        raise ValueError("as_gt_graph for nodes not implemented yet")


@as_gt_graph.register(dict)
def _dict(x, directed=True, node_key='name') -> gt.Graph:
    """Convert dict containing nodes and edges, and returns a graph-tool graph."""
    if guess_dict_type(x) == 'node_edge':

        # Check if one of the value is None
        if any([v is None for v in x.values()]):
            x = [v for v in x.values() if v is not None][0]
            if guess_df_type(x) == 'EdgeDataFrame':
                nodes = _extract_nodes(x)
                x = {"nodes": nodes, "edges": x}
                # edges = _as_graph_edge_df(x)

        nodes, edges = _as_graph_node_edge(x, node_key=node_key)

        # Create Graph
        g = gt.Graph(directed=directed) >> activate("nodes")
        g.add_edge_list(edges[["source", "target"]].to_numpy())

        # Add node metadata
        nodecols = nodes.columns
        if len(nodecols) == 1:
            augment_prop(g,  NodeDataFrame(nodes), prop_name=nodecols[0])
        elif len(nodecols) > 1:
            [augment_prop(g, NodeDataFrame(nodes), prop_name=c) for c in nodecols]

        # Add edge metadata
        g = activate(g, "edges")
        edgecols = edges.iloc[:, 2::].columns
        if len(edgecols) == 1:
            augment_prop(g,  x=EdgeDataFrame(edges), prop_name=edgecols[0])
        elif len(edgecols) > 1:
            [augment_prop(g, x=EdgeDataFrame(edges), prop_name=c) for c in edgecols]

        return activate(g, "nodes")
    else:
        raise ValueError("Other types not implemented yet")


def _as_graph_node_edge(x: Dict[pd.DataFrame, pd.DataFrame], node_key: str = 'name'):
    """Prep and check that list of nodes and edges is a proper graph_node_edge format"""
    nodes = [v[1] for v in x.items() if v[0] in ['nodes', 'vertices']][0]
    edges = [v[1] for v in x.items() if v[0] in ['edges', 'links']][0]

    assert_nodes_edges_equal(nodes, edges, node_key), "Is `name` the node key?"

    # Making sure that node_key is in front of everything and of type string
    nodes = nodes.reorder_columns([f'{node_key}'])
    nodes[f"{node_key}"] = nodes[f"{node_key}"].astype(str)

    edges.columns = edges.columns.str.lower()
    check_column(edges, column_names=["source", "target"], present=True)
    edges = _indexify_edges(edges=edges, nodes=nodes, node_key=node_key)
    return nodes, edges


def _indexify_edges(
    edges: pd.DataFrame,
    nodes: pd.DataFrame,
    node_key: str
) -> pd.DataFrame:
    """
    Creates edges data frame with `source` and `target` labelled by index in
    nodes data frame. If "name" not in nodes data frame, we use the first col
    as name. If no node data frame, we create one out of edge dataframe

    :return: edge dataframe 
    """

    nodes = nodes.copy()

    # Prep nodes and edges to be merged
    nodes['index'] = nodes.index
    edges["source"] = edges.source.astype(str)
    edges["target"] = edges.target.astype(str)

    nodes = nodes.drop_duplicates(f'{node_key}')\
                 .assign(name=lambda x: x[f'{node_key}'].astype(str))

    # Keeping track of the original col order
    cols = edges.columns

    # Using node index to name source target in graph object
    assert_nodes_mergeable(nodes, edges, left_key="source", right_key=f"{node_key}")
    assert_nodes_mergeable(nodes, edges, left_key="target", right_key=f"{node_key}")

    edges = edges.merge(nodes, how="left", left_on="source", right_on=f"{node_key}")\
                 .drop("source", axis=1)\
                 .rename(columns={"index": "source"})
    edges = edges.merge(nodes, how="left", left_on="target", right_on=f"{node_key}")\
                 .drop("target", axis=1)\
                 .rename(columns={"index": "target"})

    return edges.drop(columns=[f"{node_key}_x", f"{node_key}_y"])[cols]

@Pipeable(try_normal_call_first=True)
def summary(G):
    if G.gp.active == 'nodes':
        print(as_data_frame(G))
        G = activate(G, "edges")
        print(as_data_frame(G))
        G = activate(G, "nodes")
    elif G.gp.active == 'edges':
        print(as_data_frame(G))
        G = activate(G, "nodes")
        print(as_data_frame(G))
        G = activate(G, "edges")



"""Calculate metrics on graph"""

import graph_tool.all as gt

from .nodedataframe import NodeDataFrame
from .context import expect_nodes


def graph_component_count(G: gt.Graph, 
                          directed: bool = False) -> NodeDataFrame:
  expect_nodes(G)
  counted_comp, _ = gt.label_components(G, directed=directed)
  return NodeDataFrame({"cc": list(counted_comp)})["cc"]


def graph_largest_component(G: gt.Graph, 
                            directed: bool = False) -> NodeDataFrame:
  expect_nodes(G)
  largest_comp = gt.label_largest_component(G, directed=directed)
  return NodeDataFrame({"lc": list(largest_comp)})["lc"]



"""Graphtool to Networkx graph"""

from typing import Union, Optional

import graph_tool.all as gt
import networkx as nx

from .context import activate
from .gt_graph import as_gt_graph, _indexify_edges
from .as_data_frame import as_data_frame
from .utils import check_column

@as_gt_graph.register(nx.classes.graph.Graph)
def _networkx_graph(x, directed=True):
    """Convert networkx graph, and returns a graph-tool graph."""
    edges = networkx_to_df(x)
    g = gt.Graph(directed=directed)
    activate(g, "nodes")
    g.add_edge_list(edges[["source","target"]].to_numpy())
    return g


def networkx_to_df(G):
    # !TODO: check if nx is installed
    edges = nx.to_pandas_edgelist(G)
    check_column(edges, column_names=["source", "target"], present=True)
    return  _indexify_edges(edges)


def as_networkx(G: gt.Graph,
                directed: bool = False,
                prop_name: Optional[str] = None) -> Union[nx.Graph, nx.DiGraph]:
    activate(G, "nodes")
    nodes = as_data_frame(G)
    activate(G, "edges")
    edges = as_data_frame(G)
    new_g = nx.from_pandas_edgelist(edges)
    if directed is True:
        new_g = new_g.to_directed()

    attr = [(i, dict(group=gr)) for i,gr in enumerate(nodes[f"{prop_name}"])]
    new_g.add_nodes_from(attr)
    return new_g



"""Graph to dataframe functions"""

from tidygraphtool.nodedataframe import NodeDataFrame
from tidygraphtool.edgedataframe import EdgeDataFrame
from .context import expect_edges, expect_nodes
import pandas as pd
from pipey import Pipeable
import graph_tool.all as gt 

@Pipeable(try_normal_call_first=True)
def as_data_frame(G: gt.Graph) -> pd.DataFrame:
    if G.gp.active == 'nodes':
        return _nodes2dataframe(G)
    elif G.gp.active == 'edges':
        return _edges2dataframe(G)
    else:
        raise ValueError("Nodes or edges must be active")

def _edges2dataframe(G: gt.Graph) -> EdgeDataFrame:
    """Takes a Graph-tool graph object and returns edges data frame.
    
    Edges will always correspond to node index.
    """
    expect_edges(G)
    edges_df = EdgeDataFrame(list(G.edges()), columns=["source", "target"])\
                 .assign(source = lambda x: x.source.astype(int),
                         target = lambda x: x.target.astype(int))
    
    edgecols = list(G.ep)
    if len(edgecols) >= 1:
        if len(edgecols) == 1:
            edges_meta = EdgeDataFrame({f"{edgecols[0]}": list(G.ep[f"{edgecols[0]}"])})
        elif len(edgecols) > 1:
            edges_meta = pd.concat([EdgeDataFrame({f"{edgecols[i]}": list(G.ep[f"{edgecols[i]}"])})
                                    for i in range(len(edgecols))], axis=1)
        return pd.concat([edges_df, edges_meta], axis=1)
    else:
        return edges_df


def _nodes2dataframe(G: gt.Graph) -> NodeDataFrame:
    """
    Takes a Graph-tool graph object and returns nodes data frame.
    """
    expect_nodes(G)
    if len(G.vp) != 0:
        prop_dfs = []
        for prop in G.vp:
            prop_dfs.append(NodeDataFrame({f"{prop}": list(G.vp[f"{prop}"])}))
        prop_dfs = pd.concat(prop_dfs, axis=1)

        main_vprop = list(G.vp)[0]
        if 'name' in prop_dfs.columns and 'name' == main_vprop:
            # Making sure that name var does not conflict with anything
            name_convertible_to_int = all(prop_dfs.name.str.isdigit())
            if name_convertible_to_int == False:
                prop_dfs = prop_dfs.rename(columns={"name":"label"})
                prop_dfs["name"] = prop_dfs.index
                prop_dfs["name"] = prop_dfs["name"].astype(str)
            
            else:
                if all(prop_dfs.name.astype(int) != prop_dfs.index):
                    prop_dfs = prop_dfs.rename(columns={"name":"label"})
                    prop_dfs["name"] = prop_dfs.index
                    prop_dfs["name"] = prop_dfs["name"].astype(str)

            other_cols = list(prop_dfs.loc[:, prop_dfs.columns != 'name'])
            return prop_dfs.reorder_columns(["name"] + other_cols)
        else: 
            other_cols = list(prop_dfs.loc[:, prop_dfs.columns != f'{main_vprop}'])
            return prop_dfs.reorder_columns([f"{main_vprop}"] + other_cols)
    else:
        return NodeDataFrame({"name":list(G.vertices())})



"""Node extension for dataframe"""

from pandas import DataFrame, Series

class NodeDataFrame(DataFrame): 
    @property
    def _constructor(self): 
        return NodeDataFrame 
      
    @property
    def _constructor_sliced(self): 
        return NodeSeries

class NodeSeries(Series): 
    @property
    def _constructor(self): 
        return NodeSeries 
      
    @property
    def _constructor_sliced(self): 
        return NodeSeries 


from .as_data_frame import *
from .augment import *
from .centrality import *
from .context import *
from .edgedataframe import *
from .graph_measures import *
from .group import *
from .gt_graph import *
# from .networkx import *
from .node import *
from .nodedataframe import *
from .verbs import *
from .play import *
from .state import *


"""Active graph in particular states"""

import graph_tool.all as gt
from pipey import Pipeable

@Pipeable(try_normal_call_first=True)
def activate(G: gt.Graph, what: str) -> gt.Graph:
  """Activate context of graph object to be in node or edges modes. 
  
  Sugary syntax that builds on top of internal graph property maps.
  """
  G = G.copy()
  if what == 'nodes' or what == 'edges':
    gprop = G.new_gp('string')
    G.gp["active"] = gprop
    G.gp["active"] = f'{what}' 
    return G
  else:
    raise ValueError("Can only activate nodes or edges")


def expect_nodes(G: gt.Graph):
  assert G.gp.active == 'nodes', 'This call requires nodes to be active'


def expect_edges(G: gt.Graph):
  assert G.gp.active == 'edges', 'This call requires edges to be active'




"""Calculate metrics on nodes"""

import graph_tool.all as gt 

from pipey import Pipeable
from .nodedataframe import NodeDataFrame, NodeSeries
from .context import expect_nodes


def node_coreness(G: gt.Graph) -> NodeSeries:
    """
    Measures coreness of each node. See ?gt.kcore_decomposition
    
    Example
    =======
    .. code-block:: python
        
        mutate(g, "coreness", node_coreness(g))
    """
    expect_nodes(G)
    return NodeDataFrame({"nc": list(gt.kcore_decomposition(G))})["nc"]


def node_largest_component(G: gt.Graph) -> NodeSeries:
    """
    Label the largest component in the graph. See ?gt.label_largest_component
    
    Example
    =======
    .. code-block:: python
        
        mutate(g, "coreness", node_coreness(g))
    """
    expect_nodes(G)
    return NodeDataFrame({"lc": list(gt.label_largest_component(G))})["lc"]


"""Functions to augment graph with metdata in dataframe"""

from typing import Optional, Union
import graph_tool.all as gt
import pandas as pd
from pipey import Pipeable

from .as_data_frame import as_data_frame
from .edgedataframe import EdgeDataFrame, EdgeSeries
from .nodedataframe import NodeDataFrame, NodeSeries
from .context import expect_edges, expect_nodes
from .utils import (
    assert_edges_edges_equal,
    guess_df_type,
    check_null_values,
    _convert_types2gt,
    assert_nodes_nodes_equal,
    assert_edges_edges_equal,
    assert_index_reset
)

@Pipeable(try_normal_call_first=True)
def augment_prop(
    G: gt.Graph,
    x: pd.DataFrame,
    prop_name: Optional[str] = None
) -> gt.Graph:
    """
    Augment nodes in Graph with additional nodes metadata.

    As opposed to mutate, augment_prop modifies in-place the graph.

    :param G: A `gt.Graph` object.
    :param x: A `data Frame` containing information about the nodes in the G. 
    :param prop_name: String that matches the colname. 
    :return: A `Graph_tool` object
    """

    if isinstance(x, (NodeDataFrame, NodeSeries)):
        return _augment_prop_nodes(G, x, prop_name)
    elif isinstance(x, (EdgeDataFrame, EdgeSeries)):
        return _augment_prop_edges(G, x, prop_name)
    elif guess_df_type(x) == 'NodeDataFrame':
        return _augment_prop_nodes(G, x, prop_name)
    elif guess_df_type(x) == 'EdgeDataFrame':
        return _augment_prop_edges(G, x, prop_name)
    else:
        raise ValueError("Unknown format")


def _augment_prop_nodes(G: gt.Graph,
                        nodes: Union[NodeDataFrame, NodeSeries],
                        prop_name: Optional[str] = None) -> pd.DataFrame:
    expect_nodes(G)
    nodes = NodeDataFrame(nodes)

    if prop_name is None:
        prop_name = nodes.columns[0]

    # check_column(nodes, column_names = [prop_name], present=True)
    check_null_values(nodes, [prop_name])

    # Create internal properties maps
    prop_type = _convert_types2gt(nodes, prop_name)
    vprop = G.new_vp(f"{prop_type}")
    G.vp[f"{prop_name}"] = vprop

    assert_nodes_nodes_equal(G, nodes)
    assert_index_reset(nodes)

    main_vprop = list(G.vp)[0]
    if main_vprop in nodes.columns:
        if all([len(_) == 0 for _ in G.vp[f'{main_vprop}']]) == False:
            nodes = _reorder_nodes_like_g(G, main_vprop, nodes)

    for i in range(len(nodes)):
        # Augment graph with nodes metadata
        vprop[i] = nodes[f"{prop_name}"][i]

    return G


def _reorder_nodes_like_g(G, prop, nodes):
    vprop_vals = list(G.vp[f"{prop}"])
    
    # Finding col with same vals than main vprop
    x = [c for c in nodes.columns if set(nodes[f"{c}"]) == set(vprop_vals)]

    if len(x) == 0:
        try:
            vprop_vals = list(G.iter_vertices())
            x = [c for c in nodes.columns if set(nodes[f"{c}"]) == set(vprop_vals)]
        except:
            raise ValueError("No cols in nodes have same values than in G's main vprop")

    nodes = nodes.set_index(nodes[f'{x[0]}'])
    nodes = nodes.loc[vprop_vals]
    nodes = nodes.reset_index(drop=True)
    return nodes


def _augment_prop_edges(G: gt.Graph,
                        edges: Union[EdgeDataFrame, EdgeSeries],
                        prop_name: str) -> pd.DataFrame:
    expect_edges(G)
    edges = EdgeDataFrame(edges)

    # check_column(edges, column_names = [prop_name])
    check_null_values(edges, [prop_name])

    if 'source' not in edges.columns:
        # G = activate(G, "edges")
        edges_df = as_data_frame(G)
        edges = pd.concat([edges_df, edges], axis=1)

    # Create internal properties maps
    prop_type = _convert_types2gt(edges, prop_name)
    eprop = G.new_ep(f"{prop_type}")
    G.edge_properties[f"{prop_name}"] = eprop

    assert_edges_edges_equal(G, edges)
    assert_index_reset(edges)

    for i in range(len(edges)):
        # Augment graph with edge metadata
        e = G.edge(edges.source[i], edges.target[i])
        eprop[e] = edges.loc[i, f"{prop_name}"]

    return G



"""Popular networks simulation"""

from math import ceil

import numpy as np
import graph_tool.all as gt
from pipey import Pipeable

from .context import activate

@Pipeable(try_normal_call_first=True)
def play_sbm(n_k: int = 100,
             k: int = 3,
             p_rs: float = 0.001,
             directed: bool = True,
             self_loops: bool = False) -> gt.Graph:
             """Play SBM"""
             n_rs = ceil((p_rs * (n_k * k)) / (k - 1))
             n_rr = (1-p_rs) * (n_k * k)

             prob_mat = np.full((k,k), n_rs)
             np.fill_diagonal(prob_mat, n_rr)

             n = prob_mat.sum()
             membership = np.repeat(range(k), n/k)

             g = gt.generate_sbm(
                     b=membership, 
                     probs=prob_mat,
                     directed=directed)
             if self_loops == True:
                 gt.remove_self_loops(g)

             g = gt.extract_largest_component(g, prune=True, directed=False)

             vprop=g.new_vp("string")
             g.vp["name"] = vprop
             
             for i in range(len(list(g.vertices()))):
               vprop[i] = g.vertex(i)


             return activate(g, "nodes")




"""Verbs for network wranling"""

import re
import graph_tool.all as gt
import pandas as pd
import numpy as np

from pipey import Pipeable

from .augment import augment_prop
from .as_data_frame import as_data_frame
from .utils import check_column, check_args_order
from .context import expect_edges, expect_nodes
from .nodedataframe import NodeDataFrame
from .node import node_largest_component


#!TODO: Not working with function that takes as extra args EdgePropertyMap.
#       The reason seems that we cannot properly unpack EdgePropertyMap value.
@Pipeable(try_normal_call_first=True)
def add_property(G, colname:str = None, *args, **kwargs) -> gt.Graph:
    """
    Creates a new column based on a function. 

    Functional usage syntax:

    .. code-block:: python
        
        g = play_sbm()
        add_property(g, "degree", centrality_degree)
        summary(g)
    
    Pipeable usage syntax with extra args:

    .. code-block:: python

      (play_sbm() >>
        activate("nodes") >>
        add_property("degree_tot", centrality_degree, mode="total"))


    :param G: A gt.Graph object you wish to add property.
    :param column_name: A string that provide the name of the property.
    :param args: implicitly, we expect a callable that calcuate the desired property.
    :param kwargs: named arguments that correspond to callable extra args.
    :returns: A gt.Graph object augmented with labelled property.
    """
    G = G.copy()
    f = args[0]

    check_args_order(G, colname, f)
    if len(kwargs) == 0:
        df = f(G).rename(f"{colname}")
    else:
        df = f(G, **kwargs).rename(f"{colname}")

    return augment_prop(G, df, prop_name=colname)


#!TODO: Only works with group_hsbm, not generally, as we cannot pass colnames.
# @Pipeable(try_normal_call_first=True)
# def add_properties(G: gt.Graph, 
#                    func: Callable[[gt.Graph], pd.Series]) -> gt.Graph:
#     """Creates a new column here based on a function."""
#     G = G.copy()
#     df = func

#     [augment_prop(G, df, f"{c}") for c in df.columns]
    
#     return G

@Pipeable(try_normal_call_first=True)
def arrange(G: gt.Graph, *args, **kwargs) -> pd.DataFrame:
    """
    Arrange the graph by the values of queried columns
    
    Pipeable usage syntax:

    `.. code-block:: python
    
        (gt.collection.data["lesmis"] >> 
            activate("nodes") >> 
            add_property("degree", centrality_degree, mode="total") >>
            add_property("bet", centrality_betweenness) >> 
            arrange(["degree", "bet"], ascending=False))

    :param G: A gt.Graph object you wish to add property.
    :param args: the selected columns. 
    :param kwargs: named arguments that correspond to callable extra args, e.g. ascending=False
    :returns: A pd.DataFrame object arrange after the values of the selected column.
    """
    df_tmp = as_data_frame(G)
    
    if len(*args) > 1:
        [check_column(df_tmp, val) for val in args]
    else:
        check_column(df_tmp, [*args])
        
    return df_tmp.sort_values(*args, **kwargs)


@Pipeable(try_normal_call_first=True)
def filter_largest_component(G:gt.Graph) -> gt.Graph:
    """Extract largest component"""
    expect_nodes(G)
    G = add_property(G, "lc", node_largest_component)
    G = filter_on(G, "lc == 1")
    return G


@Pipeable(try_normal_call_first=True)
def filter_on(G: gt.Graph, criteria: str) -> gt.Graph:
    """
    Filter tidystyle on a particular criteria.
    
    Pipeable usage syntax:

    .. code-block:: python

        (gt.collection.data["lesmis"] >> 
           activate("nodes") >> 
           add_property("degree", centrality_degree, mode="total") >>
           filter_on("degree >= 10") >>
           summary())

    """
    df = as_data_frame(G)
    # _check_col_criteria(df, criteria)
    df_tmp = df.query(criteria)

    if G.gp.active == "nodes":
        expect_nodes(G)
        df["bp"] = np.where(df.iloc[:,0].isin(df_tmp.iloc[:,0]), True, False)
        G = augment_prop(G, df, prop_name="bp")
        G = gt.GraphView(G, vfilt=G.vp.bp)
        G = gt.Graph(G, prune=True)
        del G.properties[("v", "bp")]
        return G
    elif G.gp.active == "edges":
        expect_edges(G)
        df["bp"] = np.where(df["source"].isin(df_tmp["source"]) & \
            df["target"].isin(df_tmp["target"]), True, False)
        G = augment_prop(G, df, prop_name="bp")
        G = gt.GraphView(G, efilt=G.ep.bp)
        G = gt.Graph(G, prune=True)
        del G.properties[("e", "bp")]
        return G
    else:
        raise ValueError("Context must be activated to nodes or edges")


def _check_col_criteria(df, criteria):
    all_comp = re.findall("\w+", criteria)
    cols = [_ for _ in all_comp if re.match("[a-zA-Z]", _)]
    [check_column(df, [c]) for c in cols]


@Pipeable(try_normal_call_first=True)
def left_join(G: gt.Graph, 
              y: pd.DataFrame, 
              on: str = None
) -> gt.Graph:
    """
    Left join dataframe on corresponding nodes or edges in graph.
    
    IMPORTANT
    =========
    left and right key must be of the name. 
    If necessary, rename columns in y beforehand.

    """
    expect_nodes(G)
    df = NodeDataFrame(as_data_frame(G))
    df = df.merge(y, how="left", on=on)
    colnames = df.columns
    [augment_prop(G, x=df, prop_name=c) for c in colnames]
    return G


@Pipeable(try_normal_call_first=True)
def rename(G: gt.Graph, old_column_name: str, new_column_name: str) -> gt.Graph:
    """
    Rename nodes or edges property.
    
    Pipeable usage syntax:

    .. code-block:: python
    
        (gt.collection.data["lesmis"] >> 
            activate("nodes") >> 
            rename("label", "new_names") >> 
            as_data_frame())
    
    """
    G = G.copy()
    
    df = as_data_frame(G).loc[:, [old_column_name]]
    check_column(df, [old_column_name])
    df = df.rename(columns={old_column_name: new_column_name})

    if G.gp.active == 'nodes':
        del G.vp[f"{old_column_name}"]
    else:
        del G.vp[f"{old_column_name}"]

    augment_prop(G, df, new_column_name)

    return G


@Pipeable(try_normal_call_first=True)
def simplify_edges(G:gt.Graph, 
                   remove_directed: bool = True,
                   remove_parallel: bool = True, 
                   remove_loop: bool = True) -> gt.Graph:
    """Remove parallel and directed edges and self loops"""
    G = G.copy()
    if remove_parallel and remove_loop and remove_directed:
        G.set_directed(False)
        gt.remove_self_loops(G)
        gt.remove_parallel_edges(G)
    elif remove_parallel and remove_directed and not remove_loop:
        G.set_directed(False)
        gt.remove_parallel_edges(G)
    elif remove_parallel and not remove_directed and not remove_loop:
        gt.remove_parallel_edges(G)
    elif remove_loop and not remove_parallel and not remove_directed:
        gt.remove_self_loops(G)
    else:
        raise ValueError("Not implemented yet")
    return G



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='tidygraphtool',
    url='https://github.com/jstonge/tidygraphtool',
    author='Jonathan St-Onge',
    author_email='jonathanstonge7@gmail.com',
    packages=['tidygraphtool'],
    install_requires=['pipey'],
    version='0.2',
    package_dir={'': 'src'},
    license='MIT',
    long_description=long_description,
    long_description_content_type="text/markdown",
    description='A tidy API for network manipulation with Graph-tool inspired by tidygraph.',
)
